import argparse
import logging
import os
import sys


def get_args(args):
    """
    Parse command line arguments.
    """

    parser = argparse.ArgumentParser(
        description="""Serve a Markdown folder as a web blog.

The folder should have the following structure:

.
  -> config.yaml [recommended]
  -> markdown
    -> article-1.md
    -> article-2.md
    -> ...
  -> img [recommended]
    -> favicon.ico
    -> icon.png
    -> image-1.png
    -> image-2.png
    -> ...

""",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "dir",
        nargs="?",
        default=None,
        help="Base path for the blog (default: current directory)",
    )
    parser.add_argument(
        "--config",
        dest="config",
        default="config.yaml",
        required=False,
        help="Path to a configuration file (default: config.yaml in the blog root directory)",
    )
    parser.add_argument(
        "--host",
        dest="host",
        required=False,
        default="0.0.0.0",
        help="Bind host/address",
    )
    parser.add_argument(
        "--port",
        dest="port",
        required=False,
        type=int,
        default=8000,
        help="Bind port (default: 8000)",
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        required=False,
        action="store_true",
        default=False,
        help="Enable debug mode (default: False)",
    )

    return parser.parse_known_args(args)


def run():
    """
    Run the application.
    """

    from .config import init_config

    opts, _ = get_args(sys.argv[1:])
    blog_dir = opts.dir or "."
    config_file = opts.config if opts.config else os.path.join(blog_dir, "config.yaml")
    config = init_config(config_file=config_file, args=opts)
    logging.basicConfig(
        level=logging.DEBUG if config.debug else logging.INFO,
        format="%(levelname)s: %(message)s",
    )

    from .app import app

    try:
        app.start()
        app.run(host=config.host, port=config.port, debug=config.debug)
    finally:
        app.stop()


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod
from datetime import datetime, date, timezone
from typing import Any, Iterable, Tuple


def _normalize_dt(dt: datetime | date | str | None) -> float:
    if not dt:
        return 0

    if isinstance(dt, str):
        dt = datetime.fromisoformat(dt.strip())
    if isinstance(dt, date):
        dt = datetime.combine(dt, datetime.min.time())
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)

    return dt.timestamp()


class PagesSorter(ABC):
    _default_published = datetime.fromtimestamp(0)

    def __init__(self, pages: Iterable[dict]):
        self.pages = pages

    @abstractmethod
    def __call__(self, page: dict) -> Any:
        raise NotImplemented()


class PagesSortByTime(PagesSorter):
    def __call__(self, page: dict) -> float:
        return _normalize_dt(page.get("published", self._default_published))


class PagesSortByFolderAndTime(PagesSorter):
    def __call__(self, page: dict) -> Tuple:
        return (
            page.get("folder"),
            _normalize_dt(
                date.today() - page.get("published", self._default_published)
            ),
        )


class PagesSortByTimeGroupedByFolder(PagesSorter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        st = {}
        for page in self.pages:
            folder = page.get("folder", "")
            published = _normalize_dt(page.get("published", self._default_published))
            st[folder] = st.get(folder, published)
            st[folder] = max(st[folder], published)

        self._max_date_by_folder = st

    def __call__(self, page: dict) -> Tuple:
        return (
            self._max_date_by_folder[page.get("folder", "")],
            _normalize_dt(page.get("published", self._default_published)),
        )



from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timezone
from logging import getLogger
from threading import RLock
from typing import Collection

import feedparser

from ._model import Feed

feeds_lock = RLock()
logger = getLogger(__name__)


class FeedParser:
    """
    A thread-safe concurrent feedparser.

    :param urls: A list of URLs to parse
    :param cache_expiry_secs: The number of seconds to cache parsed feeds
        (default: no caching)
    """

    def __init__(
        self, urls: Collection[str] = tuple(), cache_expiry_secs: int = 0
    ) -> None:
        self.urls = urls
        self._cache_expiry_secs = cache_expiry_secs
        self._feeds_cache: dict[str, Feed] = {}

    def parse_feed(self, url: str) -> Feed | None:
        cached_feed = self._feeds_cache.get(url)
        if cached_feed and (
            not self._cache_expiry_secs
            or (
                cached_feed.last_fetched
                and (
                    datetime.now(timezone.utc) - cached_feed.last_fetched
                ).total_seconds()
                < self._cache_expiry_secs
            )
        ):
            return cached_feed

        try:
            feed = Feed.build(feedparser.parse(url))
            feed.last_fetched = datetime.now(timezone.utc)
            self._feeds_cache[url] = feed
            return feed
        except Exception:
            logger.exception("Failed to parse feed %s", url)
            return None

    def parse_feeds(self) -> dict[str, Feed]:
        """
        Parse all feeds.

        :return: A dictionary of parsed feeds in the form {href: feed}
        """
        with feeds_lock, ThreadPoolExecutor(max_workers=10) as executor:
            return {
                feed.href: feed
                for feed in executor.map(self.parse_feed, self.urls)
                if feed
            }



import datetime
import os
import re
from pathlib import Path
from typing import IO, List, Optional, Tuple, Type
from urllib.parse import urlparse

from flask import Flask, Response, abort, make_response, render_template
from markdown import markdown
from webmentions import WebmentionDirection, WebmentionsHandler
from webmentions.storage.adapters.file import FileSystemMonitor
from webmentions.server.adapters.flask import bind_webmentions

from .config import config
from .feeds import FeedAuthor, FeedParser
from .latex import MarkdownLatex
from .notifications import SmtpConfig, build_webmention_email_notifier
from .storage.mentions import FileWebmentionsStorage
from ._sorters import PagesSorter, PagesSortByTime


class BlogApp(Flask):
    """
    The main application class.
    """

    _title_header_regex = re.compile(r"^#\s*((\[(.*)])|(.*))")
    _author_regex = re.compile(r"^(.+?)\s+<([^>]+)>$")
    _url_regex = re.compile(r"^(https?:\/\/)?[\w\.\-]+\.[a-z]{2,6}\/?")
    _email_regex = re.compile(r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, template_folder=config.templates_dir, **kwargs)
        self.pages_dir = (
            Path(Path(config.content_dir) / "markdown").expanduser().resolve()
        )
        self.img_dir = config.default_img_dir
        self.css_dir = config.default_css_dir
        self.js_dir = config.default_js_dir
        self.fonts_dir = config.default_fonts_dir
        self._feed_parser = FeedParser(config.external_feeds)

        if not os.path.isdir(self.pages_dir):
            # If the `markdown` subfolder does not exist, then the whole
            # `config.content_dir` is treated as the root for Markdown files.
            self.pages_dir = config.content_dir

        img_dir = os.path.join(config.content_dir, "img")
        if os.path.isdir(img_dir):
            self.img_dir = os.path.abspath(img_dir)
        else:
            self.img_dir = config.content_dir

        css_dir = os.path.join(config.content_dir, "css")
        if os.path.isdir(css_dir):
            self.css_dir = os.path.abspath(css_dir)

        js_dir = os.path.join(config.content_dir, "js")
        if os.path.isdir(js_dir):
            self.js_dir = os.path.abspath(js_dir)

        fonts_dir = os.path.join(config.content_dir, "fonts")
        if os.path.isdir(fonts_dir):
            self.fonts_dir = os.path.abspath(fonts_dir)

        templates_dir = os.path.join(config.content_dir, "templates")
        if os.path.isdir(templates_dir):
            self.template_folder = os.path.abspath(templates_dir)

        self._init_webmentions()

    def _init_webmentions(self):
        from . import __version__

        self.mentions_dir = (
            Path(Path(config.content_dir) / "mentions").expanduser().resolve()
        )

        self.webmentions_storage = FileWebmentionsStorage(
            content_dir=self.pages_dir,
            mentions_dir=self.mentions_dir,
            base_url=config.link,
            webmentions_hard_delete=config.webmentions_hard_delete,
        )

        on_mention_processed = None
        if config.author_email and config.smtp_server:
            on_mention_processed = build_webmention_email_notifier(
                recipient=config.author_email,
                blog_base_url=config.link,
                smtp=SmtpConfig(
                    server=config.smtp_server,
                    port=config.smtp_port,
                    username=config.smtp_username,
                    password=config.smtp_password,
                    starttls=config.smtp_starttls,
                    enable_starttls_auto=config.smtp_enable_starttls_auto,
                    sender=config.smtp_sender,
                ),
            )

        self.webmentions_handler = WebmentionsHandler(
            storage=self.webmentions_storage,
            base_url=config.link,
            user_agent=f"Madblog/{__version__} ({config.link})",
            on_mention_processed=on_mention_processed,
        )

        self.filesystem_monitor = FileSystemMonitor(
            root_dir=str(self.pages_dir),
            handler=self.webmentions_handler,
            file_to_url_mapper=self._file_to_url,
            throttle_seconds=config.throttle_seconds_on_update,
        )

        if config.enable_webmentions:
            bind_webmentions(self, self.webmentions_handler)

    def _file_to_url(self, f: str) -> str:
        # Return the path relative to self.pages_dir and strip the extension
        f = os.path.relpath(f, self.pages_dir).rsplit(".", 1)[0]
        return f"{config.link}/article/{f}"

    def start(self) -> None:
        if config.enable_webmentions:
            self.filesystem_monitor.start()

    def stop(self) -> None:
        if config.enable_webmentions:
            self.filesystem_monitor.stop()

    @staticmethod
    def _parse_metadata_from_markdown(handle: IO, page: str) -> dict:
        metadata: dict = {"uri": "/article/" + page[:-3]}
        for line in handle:
            if not line:
                continue

            if not (m := re.match(r"^\[//]: # \(([^:]+):\s*(.*)\)\s*$", line)):
                break

            if m.group(1) == "published":
                metadata[m.group(1)] = datetime.datetime.fromisoformat(
                    m.group(2)
                ).date()
            else:
                metadata[m.group(1)] = m.group(2)

        return metadata

    @staticmethod
    def _infer_title_and_url_from_markdown(handle: IO) -> Tuple[str, str]:
        for line in handle:
            if not line:
                continue

            if not (m := re.match(r"^#\s+(\[?([^]]+)\]?(\((.*)\))?)\s*$", line)):
                break

            return m.group(2), m.group(4)

        return "", ""

    def _parse_page_metadata(self, page: str) -> dict:
        """
        Parse the metadata from a Markdown page
        """
        if not page.endswith(".md"):
            page = page + ".md"

        md_file = os.path.join(self.pages_dir, page)
        if not os.path.isfile(md_file):
            abort(404)

        metadata = {}
        with open(md_file, "r") as f:
            metadata.update(self._parse_metadata_from_markdown(f, page))

        metadata["title_inferred"] = not metadata.get("title")
        if not metadata.get("title"):
            with open(md_file, "r") as f:
                metadata["title"], url = self._infer_title_and_url_from_markdown(f)

            if url:
                metadata["external_url"] = url

        if not metadata.get("title"):
            metadata["title"] = os.path.splitext(os.path.basename(md_file))[0]

        if not metadata.get("published"):
            # If the `published` header isn't available in the file,
            # infer it from the file's creation date
            metadata["published"] = datetime.date.fromtimestamp(
                os.stat(md_file).st_ctime
            )
            metadata["published_inferred"] = True

        return metadata

    @classmethod
    def _parse_author(cls, metadata: dict) -> dict:
        author = None
        author_url = None
        author_photo = None

        if metadata.get("author"):
            if match := cls._author_regex.match(metadata["author"]):
                author = match[1]
                if link := match[2].strip():
                    author_url = link
            else:
                author = metadata["author"]
        else:
            author = config.author
            author_url = config.author_url

        if author_url and cls._email_regex.match(author_url):
            author_url = "mailto:" + author_url

        if metadata.get("author_photo"):
            if link := metadata["author_photo"].strip():
                if cls._url_regex.match(link):
                    author_photo = link
        else:
            author_photo = config.author_photo

        return {
            "author": author,
            "author_url": author_url,
            "author_photo": author_photo,
        }

    def get_page(
        self,
        page: str,
        title: Optional[str] = None,
        skip_header: bool = False,
        skip_html_head: bool = False,
    ) -> Response:
        """
        Get the HTML for a Markdown page
        """
        if not page.endswith(".md"):
            page = page + ".md"

        metadata = self._parse_page_metadata(page)
        title = title or metadata.get("title") or config.title
        author_info = self._parse_author(metadata)
        mentions = self.webmentions_handler.render_webmentions(
            self.webmentions_handler.retrieve_stored_webmentions(
                config.link + metadata.get("uri", ""),
                direction=WebmentionDirection.IN,
            )
        )

        with open(os.path.join(self.pages_dir, page), "r") as f:
            content = self._parse_content(f)

        html = render_template(
            "article.html",
            config=config,
            title=title,
            uri=metadata.get("uri"),
            url=config.link + metadata.get("uri", ""),
            external_url=metadata.get("external_url"),
            image=metadata.get("image"),
            description=metadata.get("description"),
            published_datetime=metadata.get("published"),
            published=metadata["published"].strftime("%b %d, %Y"),
            content=markdown(
                content,
                extensions=["fenced_code", "codehilite", "tables", MarkdownLatex()],
            ),
            skip_header=skip_header,
            skip_html_head=skip_html_head,
            mentions=mentions,
            **author_info,
        )

        response = make_response(html)
        if config.webmention_url:
            response.headers["Link"] = f'<{config.webmention_url}>; rel="webmention"'

        return response

    @staticmethod
    def _parse_content(handle: IO) -> str:
        """
        Prepare content for rendering.
        """
        content = ""
        processed_title = False

        for line in handle:
            if line.startswith("# "):
                if not processed_title:
                    processed_title = True
                    continue

            content += line

        return content

    def _get_page_content(self, page: str, **kwargs) -> str:
        """Return the HTML content of a page as a string (not a Response)."""
        result = self.get_page(page, **kwargs)
        if isinstance(result, Response):
            return result.get_data(as_text=True)
        return str(result) if result else ""

    def _get_pages_from_files(
        self,
        *,
        with_content: bool = False,
        skip_header: bool = False,
        skip_html_head: bool = False,
    ):
        pages_dir = getattr(app, "pages_dir", "")
        assert pages_dir  # for mypy
        pages_dir = str(pages_dir).rstrip("/")
        return [
            {
                "path": os.path.join(root[len(pages_dir) + 1 :], f),
                "folder": root[len(pages_dir) + 1 :],
                "content": (
                    self._get_page_content(
                        os.path.join(root, f),
                        skip_header=skip_header,
                        skip_html_head=skip_html_head,
                    )
                    if with_content
                    else ""
                ),
                **self._parse_page_metadata(
                    os.path.join(root[len(pages_dir) + 1 :], f)
                ),
            }
            for root, _, files in os.walk(pages_dir, followlinks=True)
            for f in files
            if f.endswith(".md")
        ]

    def _get_pages_from_feeds(self, *, with_content: bool = False):
        return [
            {
                "uri": entry.link,
                "external_url": entry.link,
                "folder": "",
                "source": urlparse(entry.link).netloc,
                "source_logo": feed.logo,
                "content": entry.content if with_content else "",
                "title": entry.title,
                "description": entry.description,
                "image": entry.enclosure,
                "published": entry.published,
                "author": next(
                    (
                        author
                        for author in (
                            entry.authors
                            or feed.authors
                            or (
                                [
                                    FeedAuthor(
                                        name=config.author,
                                        uri=config.author_url or "",
                                        email="",
                                    )
                                ]
                                if config.author
                                else []
                            )
                        )
                    ),
                    None,
                ),
            }
            for feed in self._feed_parser.parse_feeds().values()
            for entry in feed.entries
        ]

    def get_pages(
        self,
        *,
        with_content: bool = False,
        skip_header: bool = False,
        skip_html_head: bool = False,
        sorter: Type[PagesSorter] = PagesSortByTime,
        reverse: bool = True,
    ) -> List[Tuple[int, dict]]:
        local_pages = self._get_pages_from_files(
            with_content=with_content,
            skip_header=skip_header,
            skip_html_head=skip_html_head,
        )

        remote_pages = self._get_pages_from_feeds(with_content=with_content)
        pages = local_pages + remote_pages
        sorter_func = sorter(pages)
        pages.sort(key=sorter_func, reverse=reverse)
        return list(enumerate(pages))


app = BlogApp(__name__)

from .routes import *

# vim:sw=4:ts=4:et:



import hashlib
import json
import logging
import re

from collections import defaultdict
from dataclasses import asdict
from datetime import datetime, timezone
from enum import Enum
from pathlib import Path
from threading import RLock
from urllib.parse import urlparse

from webmentions import (
    Webmention,
    WebmentionDirection,
    WebmentionsHandler,
    WebmentionStatus,
    WebmentionsStorage,
)

logger = logging.getLogger(__name__)


class FileWebmentionsStorage(WebmentionsStorage):
    """
    File-based storage for Webmentions.

    Stores each Webmention as a Markdown file with metadata in comments.
    """

    def __init__(
        self,
        content_dir: str | Path,
        mentions_dir: str | Path,
        *,
        base_url: str,
        webmentions_hard_delete: bool = False,
        **_,
    ):
        self.content_dir = Path(content_dir).resolve()
        self.mentions_dir = Path(mentions_dir).resolve()
        self.mentions_dir.mkdir(exist_ok=True, parents=True)
        self.base_url = base_url
        self._webmentions_hard_delete = webmentions_hard_delete
        self._resource_locks = defaultdict(RLock)
        self._watcher_lock = RLock()
        self._webmentions_handler: WebmentionsHandler | None = None

    def store_webmention(self, mention: Webmention):
        """
        Store Webmention as Markdown file
        """

        if mention.direction == WebmentionDirection.IN:
            # Extract post slug from target URL
            post_slug = self._extract_post_slug(mention.target)
            post_mentions_dir = (
                self.mentions_dir / WebmentionDirection.IN.value / post_slug
            )
        else:
            # Extract post slug from source URL
            post_slug = self._extract_post_slug(mention.source)
            post_mentions_dir = (
                self.mentions_dir / WebmentionDirection.OUT.value / post_slug
            )

        post_mentions_dir.mkdir(exist_ok=True, parents=True)

        # Generate safe filename
        filename = self._generate_mention_filename(mention.source, "webmention")
        filepath = post_mentions_dir / filename

        # Prepare metadata
        mention.created_at = datetime.now(timezone.utc)
        mention.updated_at = mention.created_at

        with self._resource_locks[filepath]:
            # Parse existing file metadata if file exists
            if filepath.exists():
                with open(filepath, "r", encoding="utf-8") as f:
                    existing_content = f.read()

                existing_metadata = self._parse_metadata(existing_content)
                if existing_metadata:
                    mention.created_at = (
                        Webmention.build(existing_metadata).created_at
                        or mention.created_at
                    )

            # Write as Markdown with YAML frontmatter
            content = self._format_webmention_markdown(mention)

            # Atomic write
            self._atomic_write(filepath, content)

        return filepath

    def retrieve_webmentions(
        self, resource: str, direction: WebmentionDirection
    ) -> list[Webmention]:
        """
        Retrieve stored Webmentions for a given resource.
        """

        post_slug = self._extract_post_slug(resource)
        post_mentions_dir = self.mentions_dir / direction.value / post_slug
        webmentions = []

        if not post_mentions_dir.exists():
            return webmentions

        for md_file in post_mentions_dir.glob("webmention-*.md"):
            try:
                with open(md_file, "r", encoding="utf-8") as f:
                    content = f.read()

                metadata = self._parse_metadata(content)
                if metadata.get("status") != WebmentionStatus.CONFIRMED.value:
                    continue

                webmentions.append(Webmention.build(metadata))
            except Exception as e:
                logger.error("Error parsing Webmention in %s: %s", md_file, e)
                continue

        return sorted(
            webmentions, key=lambda x: x.published or x.created_at, reverse=True
        )

    def delete_webmention(
        self,
        source: str,
        target: str,
        direction: WebmentionDirection,
    ):
        """Mark a stored mention as deleted by updating its metadata."""

        if direction == WebmentionDirection.IN:
            post_slug = self._extract_post_slug(target)
            post_mentions_dir = (
                self.mentions_dir / WebmentionDirection.IN.value / post_slug
            )
        else:
            post_slug = self._extract_post_slug(source)
            post_mentions_dir = (
                self.mentions_dir / WebmentionDirection.OUT.value / post_slug
            )

        filename = self._generate_mention_filename(source, "webmention")
        filepath = post_mentions_dir / filename
        if not filepath.exists():
            return None

        if self._webmentions_hard_delete:
            with self._resource_locks[filepath]:
                filepath.unlink(missing_ok=True)
            return filepath

        with self._resource_locks[filepath]:
            with open(filepath, "r", encoding="utf-8") as f:
                existing_content = f.read()

            existing_metadata = self._parse_metadata(existing_content)
            metadata = {
                **existing_metadata,
                "type": existing_metadata.get("type", "webmention"),
                "source": source,
                "target": target,
                "updated_at": datetime.now(timezone.utc),
                "status": "deleted",
            }

            content = self._format_webmention_markdown(Webmention.build(metadata))
            self._atomic_write(filepath, content)

        return filepath

    @staticmethod
    def _parse_metadata(content: str) -> dict:
        """
        Parse metadata from Markdown comments.
        """

        metadata = {}
        for line in content.splitlines():
            match = re.match(r"\[//]: # \(([^:]+): (.+)\)", line)
            if match:
                key, value = match.groups()
                metadata[key.strip()] = value.strip()

        return metadata

    @staticmethod
    def _format_webmention_markdown(mention: Webmention):
        """
        Format Webmention as Markdown and include metadata as comments.
        """

        md_content = ""

        for key, value in asdict(mention).items():
            if key == "content" or value is None:
                continue
            if isinstance(value, Enum):
                value = value.value
            if isinstance(value, datetime):
                value = value.isoformat()
            if isinstance(value, (list, dict)):
                value = json.dumps(value)

            md_content += f"[//]: # ({key}: {value})\n"

        md_content += f"\n{mention.content}\n"
        return md_content

    @staticmethod
    def _atomic_write(filepath: Path, content: str):
        """Atomic file write to prevent corruption"""
        temp_path = filepath.with_suffix(".tmp")
        try:
            with open(temp_path, "w", encoding="utf-8") as f:
                f.write(content)
            temp_path.replace(filepath)  # Atomic on most filesystems
        except Exception:
            if temp_path.exists():
                temp_path.unlink()
            raise

    @classmethod
    def _extract_post_slug(cls, target_url: str):
        """Extract post slug from target URL"""
        # Customize based on your URL structure
        # e.g., https://yourblog.com/post/hello-world -> hello-world
        parts = target_url.rstrip("/").split("/")
        return cls._safe_filename(parts[-1])

    @classmethod
    def _generate_mention_filename(
        cls, source_url: str, mention_type: str = "webmention"
    ):
        """
        Generate unique, safe filename for mentions.
        """
        url_hash = hashlib.md5(source_url.encode()).hexdigest()[:8]

        # Extract domain for readability
        domain = urlparse(source_url).netloc.replace(".", "-")
        domain = cls._safe_filename(domain)

        return f"{mention_type}-{domain}-{url_hash}.md"

    @staticmethod
    def _safe_filename(text: str, max_length: int = 50) -> str:
        """
        Generate safe filenames from Webmention source URLs.
        """
        # Remove/replace unsafe characters
        safe = re.sub(r"[^\w\s-]", "", text)
        safe = re.sub(r"[-\s]+", "-", safe)
        return safe[:max_length].strip("-")



from dataclasses import dataclass, field
from datetime import datetime


def _parse_link(data: dict) -> str:
    return (
        data.get("link")
        or next(
            iter(
                [
                    link
                    for link in data.get("links", [])
                    if link.get("rel") == "alternate"
                    and link.get("type") == "text/html"
                ]
            ),
            None,
        )
        or data.get("id", "")
    )


def _parse_links(data: dict) -> list["FeedLink"]:
    return [
        FeedLink(href=link["href"], rel=link["rel"], type=link["type"])
        for link in data.get("links", [])
    ]


def _parse_authors(data: dict) -> list["FeedAuthor"]:
    return [
        author
        for a in data.get(
            "authors",
            ([data["author_detail"]] if data.get("author_detail") else []),
        )
        if (author := FeedAuthor.build(a))
    ]


def _parse_dt(data: str | datetime | None) -> datetime | None:
    if not data:
        return None
    if isinstance(data, datetime):
        return data

    import email.utils
    import re
    from datetime import timezone

    # Normalize common timezone abbreviations to numeric offsets
    _tz_abbr = {
        "GMT": "+0000",
        "UTC": "+0000",
        "EST": "-0500",
        "EDT": "-0400",
        "CST": "-0600",
        "CDT": "-0500",
        "MST": "-0700",
        "MDT": "-0600",
        "PST": "-0800",
        "PDT": "-0700",
    }

    # 1. Try ISO 8601 first (covers most Atom feeds)
    try:
        return datetime.fromisoformat(data)
    except ValueError:
        pass

    # 2. Try email.utils.parsedate_to_datetime (handles RFC 2822 / RSS dates
    #    including timezone abbreviations like GMT)
    try:
        return email.utils.parsedate_to_datetime(data)
    except (ValueError, TypeError):
        pass

    # 3. Replace known tz abbreviations and retry strptime
    normalized = data.strip()
    for abbr, offset in _tz_abbr.items():
        if normalized.endswith(abbr):
            normalized = normalized[: -len(abbr)].rstrip() + " " + offset
            break

    _formats = [
        "%a, %d %b %Y %H:%M:%S %z",  # RFC 822
        "%d %b %Y %H:%M:%S %z",  # without weekday
        "%Y-%m-%dT%H:%M:%S%z",  # ISO without separators in tz
        "%Y-%m-%d %H:%M:%S",  # naive datetime
        "%Y-%m-%d",  # date only
    ]

    for fmt in _formats:
        try:
            return datetime.strptime(normalized, fmt)
        except ValueError:
            continue

    # 4. Last resort: strip any trailing non-numeric tz and parse as UTC
    stripped = re.sub(r"\s+[A-Z]{1,5}$", "", data.strip())
    for fmt in _formats:
        try:
            dt = datetime.strptime(stripped, fmt)
            return dt.replace(tzinfo=timezone.utc)
        except ValueError:
            continue

    return None


@dataclass
class FeedLink:
    """
    A link to a feed
    """

    href: str
    rel: str
    type: str


@dataclass
class FeedAuthor:
    """
    A feed author
    """

    name: str
    email: str
    uri: str

    @classmethod
    def build(cls, data: dict):
        ret = cls(
            name=data.get("name", ""),
            email=data.get("email", ""),
            uri=data.get("href", ""),
        )

        if not (ret.name or ret.email or ret.uri):
            return None

        return ret


@dataclass
class FeedEntry:
    """
    Feed entry model
    """

    title: str
    link: str
    content: str = ""
    description: str | None = None
    published: datetime | None = None
    authors: list[FeedAuthor] = field(default_factory=list)
    enclosure: str | None = None
    links: list[FeedLink] = field(default_factory=list)
    updated: datetime | None = None
    feed: Feed | None = None

    @classmethod
    def build(cls, data: dict):
        return cls(
            title=data.get("title", ""),
            link=_parse_link(data),
            content=next(
                iter(
                    [
                        item
                        for item in (data.get("content") or [{}])
                        if (item.get("type") or "") == "text/html"
                    ]
                ),
                {},
            ).get("value", ""),
            description=(
                data.get("summary") or data.get("summary_detail", {}).get("value")
            ),
            published=_parse_dt(data.get("published")),
            authors=_parse_authors(data),
            enclosure=next(
                iter(
                    [
                        href
                        for link in data.get("links", [])
                        if link.get("rel") == "enclosure" and (href := link.get("href"))
                    ]
                ),
                None,
            ),
            links=_parse_links(data),
            updated=_parse_dt(data.get("updated")),
        )


@dataclass
class Feed:
    """
    Feed model
    """

    title: str
    link: str
    href: str
    description: str | None = None
    language: str | None = None
    logo: str | None = None
    links: list[FeedLink] = field(default_factory=list)
    authors: list[FeedAuthor] = field(default_factory=list)
    entries: list[FeedEntry] = field(default_factory=list)
    updated: datetime | None = None
    last_fetched: datetime | None = None

    @classmethod
    def build(cls, data: dict):
        feed = data.get("feed", {})
        feed_obj = cls(
            title=feed.get("title", ""),
            link=_parse_link(data),
            href=data.get("href", ""),
            description=(
                feed.get("subtitle")
                or feed.get("description")
                or feed.get("subtitle_detail", {}).get("value")
            ),
            language=feed.get("language"),
            logo=feed.get("logo"),
            links=_parse_links(data),
            authors=_parse_authors(feed),
            entries=[FeedEntry.build(entry) for entry in data.get("entries", [])],
            updated=_parse_dt(feed.get("updated")),
        )

        for entry in feed_obj.entries:
            entry.feed = feed_obj

        return feed_obj



#!/usr/bin/env python

from setuptools import setup, find_packages


def readfile(file):
    with open(file, "r") as f:
        return f.read()


setup(
    name="madblog",
    version="0.5.4",
    author="Fabio Manganiello",
    author_email="info@fabiomanganiello.com",
    description="A minimal platform for Markdown-based blogs",
    license="MIT",
    python_requires=">= 3.8",
    keywords="blog markdown",
    url="https://git.fabiomanganiello.com/madblog",
    packages=find_packages(include=["madblog", "madblog.*"]),
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "madblog=madblog.cli:run",
        ],
    },
    long_description=readfile("README.md"),
    long_description_content_type="text/markdown",
    classifiers=[
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
    ],
    install_requires=[
        "feedgen2",
        "feedparser",
        "flask",
        "markdown",
        "pygments",
        "pyyaml",
        "requests",
        "watchdog",
        "webmentions[file]",
    ],
)



import os
import re
from argparse import Namespace
from dataclasses import dataclass, field
from typing import List, Optional

import yaml
from webmentions import WebmentionStatus


@dataclass
class Config:
    """
    Configuration for the blog
    """

    title: str = "Blog"
    description: str = ""
    link: str = "/"
    home_link: str = "/"
    external_links: List[str] = field(default_factory=list)
    host: str = "0.0.0.0"
    port: int = 8000
    language: str = "en-US"
    logo: str = "/img/icon.png"
    header: bool = True
    content_dir: str = "."
    categories: List[str] = field(default_factory=list)
    short_feed: bool = False
    max_entries_per_feed: int = 10
    enable_webmentions: bool = True
    webmentions_hard_delete: bool = False
    default_webmention_status: WebmentionStatus = WebmentionStatus.CONFIRMED
    debug: bool = False
    basedir = os.path.abspath(os.path.dirname(__file__))
    author: str | None = None
    author_url: str | None = None
    author_photo: str | None = None
    throttle_seconds_on_update: int = 10
    author_email: str | None = None
    smtp_server: str | None = None
    smtp_port: int = 587
    smtp_username: str | None = None
    smtp_password: str | None = None
    smtp_starttls: bool = True
    smtp_enable_starttls_auto: bool = True
    smtp_sender: str | None = None
    view_mode: str = "cards"  # "cards", "list", or "full"
    external_feeds: List[str] = field(default_factory=list)
    feeds_cache_expiry_secs: int = 300

    @property
    def templates_dir(self) -> str:
        return os.path.join(self.basedir, "templates")

    @property
    def static_dir(self) -> str:
        return os.path.join(self.basedir, "static")

    @property
    def default_css_dir(self) -> str:
        return os.path.join(self.static_dir, "css")

    @property
    def default_js_dir(self) -> str:
        return os.path.join(self.static_dir, "js")

    @property
    def default_fonts_dir(self) -> str:
        return os.path.join(self.static_dir, "fonts")

    @property
    def default_img_dir(self) -> str:
        return os.path.join(self.static_dir, "img")

    @property
    def webmention_url(self) -> Optional[str]:
        from flask import request

        if not self.enable_webmentions:
            return None

        return (
            f'{self.link.rstrip("/")}/webmentions'
            if re.match(r"^https?://", self.link)
            else f'{request.host_url.rstrip("/")} /webmentions'
        )


config = Config()


def _init_config_from_file(config_file: str):
    cfg = {}

    if os.path.isfile(config_file):
        with open(config_file, "r") as f:
            cfg = yaml.safe_load(f) or {}

    if cfg.get("content_dir"):
        config.content_dir = cfg["content_dir"]
    if cfg.get("title"):
        config.title = cfg["title"]
    if cfg.get("description"):
        config.description = cfg["description"]
    if cfg.get("link"):
        config.link = cfg["link"]
    if cfg.get("home_link"):
        config.home_link = cfg["home_link"]
    if cfg.get("external_links"):
        config.external_links = cfg["external_links"]
    if cfg.get("host"):
        config.host = cfg["host"]
    if cfg.get("port"):
        config.port = int(cfg["port"])
    if cfg.get("logo") is not None:
        config.logo = cfg["logo"]
    if cfg.get("language"):
        config.language = cfg["language"]
    if cfg.get("header"):
        config.header = bool(cfg["header"])
    if cfg.get("short_feed"):
        config.short_feed = bool(cfg["short_feed"])
    if cfg.get("max_entries_per_feed") is not None:
        config.max_entries_per_feed = int(cfg["max_entries_per_feed"])
    if cfg.get("enable_webmentions") is not None:
        config.enable_webmentions = bool(cfg["enable_webmentions"])
    if cfg.get("webmentions_hard_delete") is not None:
        config.webmentions_hard_delete = bool(cfg["webmentions_hard_delete"])
    if cfg.get("default_webmention_status"):
        config.default_webmention_status = WebmentionStatus(
            cfg["default_webmention_status"]
        )
    if cfg.get("debug") is not None:
        config.debug = bool(cfg["debug"])
    if cfg.get("author"):
        config.author = cfg["author"]
    if cfg.get("author_url"):
        config.author_url = cfg["author_url"]
    if cfg.get("author_photo"):
        config.author_photo = cfg["author_photo"]
    if cfg.get("throttle_seconds_on_update"):
        config.throttle_seconds_on_update = int(cfg["throttle_seconds_on_update"])
    if cfg.get("author_email"):
        config.author_email = cfg["author_email"]
    if cfg.get("smtp_server"):
        config.smtp_server = cfg["smtp_server"]
    if cfg.get("smtp_port"):
        config.smtp_port = int(cfg["smtp_port"])
    if cfg.get("smtp_username"):
        config.smtp_username = cfg["smtp_username"]
    if cfg.get("smtp_password"):
        config.smtp_password = cfg["smtp_password"]
    if cfg.get("smtp_starttls") is not None:
        config.smtp_starttls = bool(cfg["smtp_starttls"])
    if cfg.get("smtp_enable_starttls_auto") is not None:
        config.smtp_enable_starttls_auto = bool(cfg["smtp_enable_starttls_auto"])
    if cfg.get("smtp_sender"):
        config.smtp_sender = cfg["smtp_sender"]
    if cfg.get("view_mode"):
        config.view_mode = cfg["view_mode"]
    if cfg.get("external_feeds"):
        config.external_feeds = cfg["external_feeds"]
    if cfg.get("feeds_cache_expiry_secs"):
        config.feeds_cache_expiry_secs = int(cfg["feeds_cache_expiry_secs"])
    config.categories = cfg.get("categories", [])


def _init_config_from_env():
    if os.getenv("MADBLOG_TITLE"):
        config.title = os.environ["MADBLOG_TITLE"]
    if os.getenv("MADBLOG_DESCRIPTION"):
        config.description = os.environ["MADBLOG_DESCRIPTION"]
    if os.getenv("MADBLOG_LINK"):
        config.link = os.environ["MADBLOG_LINK"]
    if os.getenv("MADBLOG_HOME_LINK"):
        config.home_link = os.environ["MADBLOG_HOME_LINK"]
    if os.getenv("MADBLOG_EXTERNAL_LINKS"):
        config.external_links = re.split(
            r"[,\s]+", os.environ["MADBLOG_EXTERNAL_LINKS"].strip()
        )
    if os.getenv("MADBLOG_HOST"):
        config.host = os.environ["MADBLOG_HOST"]
    if os.getenv("MADBLOG_PORT"):
        config.port = int(os.environ["MADBLOG_PORT"])
    if os.getenv("MADBLOG_CONTENT_DIR"):
        config.content_dir = os.environ["MADBLOG_CONTENT_DIR"]
    if os.getenv("MADBLOG_LOGO"):
        config.logo = os.environ["MADBLOG_LOGO"]
    if os.getenv("MADBLOG_LANGUAGE"):
        config.language = os.environ["MADBLOG_LANGUAGE"]
    if os.getenv("MADBLOG_HEADER"):
        config.header = os.environ["MADBLOG_HEADER"] == "1"
    if os.getenv("MADBLOG_SHORT_FEED"):
        config.short_feed = os.environ["MADBLOG_SHORT_FEED"] == "1"
    if os.getenv("MADBLOG_MAX_ENTRIES_PER_FEED"):
        config.max_entries_per_feed = int(os.environ["MADBLOG_MAX_ENTRIES_PER_FEED"])
    if os.getenv("MADBLOG_ENABLE_WEBMENTIONS"):
        config.enable_webmentions = os.environ["MADBLOG_ENABLE_WEBMENTIONS"] == "1"
    if os.getenv("MADBLOG_WEBMENTIONS_HARD_DELETE"):
        config.webmentions_hard_delete = (
            os.environ["MADBLOG_WEBMENTIONS_HARD_DELETE"] == "1"
        )
    if os.getenv("MADBLOG_DEFAULT_WEBMENTION_STATUS"):
        config.default_webmention_status = WebmentionStatus(
            os.environ["MADBLOG_DEFAULT_WEBMENTION_STATUS"]
        )
    if os.getenv("MADBLOG_DEBUG"):
        config.debug = os.environ["MADBLOG_DEBUG"] == "1"
    if os.getenv("MADBLOG_VIEW_MODE"):
        config.view_mode = os.environ["MADBLOG_VIEW_MODE"]
    if os.getenv("MADBLOG_CATEGORIES"):
        config.categories = re.split(r"[,\s]+", os.environ["MADBLOG_CATEGORIES"])
    if os.getenv("MADBLOG_AUTHOR"):
        config.author = os.environ["MADBLOG_AUTHOR"]
    if os.getenv("MADBLOG_AUTHOR_URL"):
        config.author_url = os.environ["MADBLOG_AUTHOR_URL"]
    if os.getenv("MADBLOG_AUTHOR_PHOTO"):
        config.author_photo = os.environ["MADBLOG_AUTHOR_PHOTO"]
    if os.getenv("MADBLOG_THROTTLE_SECONDS_ON_UPDATE"):
        config.throttle_seconds_on_update = int(
            os.environ["MADBLOG_THROTTLE_SECONDS_ON_UPDATE"]
        )
    if os.getenv("MADBLOG_AUTHOR_EMAIL"):
        config.author_email = os.environ["MADBLOG_AUTHOR_EMAIL"]
    if os.getenv("MADBLOG_SMTP_SERVER"):
        config.smtp_server = os.environ["MADBLOG_SMTP_SERVER"]
    if os.getenv("MADBLOG_SMTP_PORT"):
        config.smtp_port = int(os.environ["MADBLOG_SMTP_PORT"])
    if os.getenv("MADBLOG_SMTP_USERNAME"):
        config.smtp_username = os.environ["MADBLOG_SMTP_USERNAME"]
    if os.getenv("MADBLOG_SMTP_PASSWORD"):
        config.smtp_password = os.environ["MADBLOG_SMTP_PASSWORD"]
    if os.getenv("MADBLOG_SMTP_STARTTLS"):
        config.smtp_starttls = os.environ["MADBLOG_SMTP_STARTTLS"] == "1"
    if os.getenv("MADBLOG_SMTP_ENABLE_STARTTLS_AUTO"):
        config.smtp_enable_starttls_auto = (
            os.environ["MADBLOG_SMTP_ENABLE_STARTTLS_AUTO"] == "1"
        )
    if os.getenv("MADBLOG_SMTP_SENDER"):
        config.smtp_sender = os.environ["MADBLOG_SMTP_SENDER"]
    if os.getenv("MADBLOG_EXTERNAL_FEEDS"):
        config.external_feeds = re.split(
            r"[,\s]+", os.environ["MADBLOG_EXTERNAL_FEEDS"].strip()
        )
    if os.getenv("MADBLOG_FEEDS_CACHE_EXPIRY_SECS"):
        config.feeds_cache_expiry_secs = int(
            os.environ["MADBLOG_FEEDS_CACHE_EXPIRY_SECS"]
        )


def _init_config_from_cli(args: Optional[Namespace]):
    if not args:
        return

    if args.dir:
        config.content_dir = args.dir
    if args.host:
        config.host = args.host
    if args.port:
        config.port = args.port
    if args.debug is not None:
        config.debug = args.debug


def init_config(
    config_file: str = "config.yaml", args: Optional[Namespace] = None
) -> Config:
    config_file = os.path.abspath(os.path.expanduser(config_file))
    _init_config_from_file(config_file)
    _init_config_from_env()
    _init_config_from_cli(args)

    # Normalize/expand paths
    config.content_dir = os.path.abspath(os.path.expanduser(config.content_dir))

    return config


# vim:sw=4:ts=4:et:



import os
import tempfile
import unittest
from pathlib import Path
from xml.etree import ElementTree


class FeedRouteTest(unittest.TestCase):
    def setUp(self):
        # Import inside setUp so tests are resilient to execution order.
        from madblog.app import app
        from madblog.config import config

        self.app = app
        self.config = config

        self._tmpdir = tempfile.TemporaryDirectory()
        self.addCleanup(self._tmpdir.cleanup)

        root = Path(self._tmpdir.name)
        markdown_dir = root / "markdown"
        markdown_dir.mkdir(parents=True, exist_ok=True)

        # Write a few posts.
        (markdown_dir / "post1.md").write_text(
            "\n".join(
                [
                    "[//]: # (title: Post 1)",
                    "[//]: # (description: Desc 1)",
                    "[//]: # (image: /img/test.png)",
                    "[//]: # (published: 2025-02-10)",
                    "",
                    "# Hello",
                    "Body 1",
                    "",
                ]
            ),
            encoding="utf-8",
        )

        (markdown_dir / "post2.md").write_text(
            "\n".join(
                [
                    "[//]: # (title: Post 2)",
                    "[//]: # (description: Desc 2)",
                    "[//]: # (published: 2025-02-11)",
                    "",
                    "# Hello",
                    "Body 2",
                    "",
                ]
            ),
            encoding="utf-8",
        )

        (markdown_dir / "post3.md").write_text(
            "\n".join(
                [
                    "[//]: # (title: Post 3)",
                    "[//]: # (description: Desc 3)",
                    "[//]: # (published: 2025-02-12)",
                    "",
                    "# Hello",
                    "Body 3",
                    "",
                ]
            ),
            encoding="utf-8",
        )

        # Point the app at the temp content.
        self.config.content_dir = str(root)
        self.config.link = "https://example.com"
        self.config.title = "Example"
        self.config.description = "Example feed"
        self.config.max_entries_per_feed = 2

        # Ensure app reads from our markdown directory.
        self.app.pages_dir = str(markdown_dir)

        # Tests shouldn't depend on webmentions.
        self.config.enable_webmentions = False

        self.client = self.app.test_client()

    def test_rss_feed_ok_and_limited(self):
        rsp = self.client.get("/feed.rss")
        self.assertEqual(rsp.status_code, 200)
        self.assertIn("application/rss+xml", rsp.content_type)

        root = ElementTree.fromstring(rsp.data)
        self.assertEqual(root.tag, "rss")

        channel = root.find("channel")
        self.assertIsNotNone(channel)
        items = channel.findall("item")
        self.assertEqual(len(items), 2)

    def test_atom_feed_ok_and_limited(self):
        rsp = self.client.get("/feed.atom")
        self.assertEqual(rsp.status_code, 200)
        self.assertIn("application/atom+xml", rsp.content_type)

        root = ElementTree.fromstring(rsp.data)
        self.assertTrue(root.tag.endswith("feed"))

        entries = root.findall("{http://www.w3.org/2005/Atom}entry")
        self.assertEqual(len(entries), 2)

    def test_short_feed_does_not_error(self):
        rsp = self.client.get("/feed.rss?short")
        self.assertEqual(rsp.status_code, 200)
        ElementTree.fromstring(rsp.data)


if __name__ == "__main__":
    unittest.main()
